using JuMP, Ipopt
using CSV, DataFrames
using LinearAlgebra
using Distributions
using Plots

function load_data()
    # feeder size
    n = 4
    e = 3

    # line parameters
    r = [0.195;0.11;0.97]
    x = [0.124;0.027;0.093]
    node_f = [0;1;2]
    node_t = [1;2;3]

    # load data 
    d_p = [-0.15;0;0.1]
    d_q = [0;0;0]

    # DER reactive power capability
    q̅ = [0.08;0.08;0.10]
    q̲ = -q̅

    # Branch-bus-incidence-matrix
    Ã = zeros(e,n)
    for j in 1:e
        Ã[j,node_f[j]+1] = 1
        Ã[j,node_t[j]+1] = -1
    end
    a_0 = Ã[:,1]
    A = Ã[:,2:end]

    # LinDistFlow sensitivities
    R = 2*inv(A)*diagm(r)*inv(A)'
    X = 2*inv(A)*diagm(x)*inv(A)'

    return Dict(:R => R, :X => X, :d_p => d_p, :d_q => d_q, :q̅ => q̅, :q̲ => q̲,
                :v̅ => 1.05^2, :v̲ => 0.95^2, :n => n, :e => e, :A => A, :Ã => Ã)
end

data = load_data()

voltage(p,q) = 1 .+ data[:R]*p .+ data[:X]*q

function droop(bus,v)
    # droop curve parameters 
    v1 = 0.95^2; v2 = 0.99^2; v3 = 1.01^2; v4 = 1.05^2; q=0;
    # piecewise linear control law
    v[bus] < v1 ? q = data[:q̅][bus] : NaN
    v1 <= v[bus] && v[bus] < v2 ? q = data[:q̅][bus] * (v2 - v[bus])/(v2 - v1) : NaN
    v2 <= v[bus] && v[bus] < v3 ? q = 0 : NaN
    v3 <= v[bus] && v[bus] < v4 ? q = data[:q̲][bus] * (v[bus] - v3)/(v4 - v3) : NaN 
    v[bus] >= v4 ? q = data[:q̲][bus] : NaN
    return q
end

function DistOPF(data,q_current)
    model = Model(optimizer_with_attributes(Ipopt.Optimizer))
    JuMP.set_silent(model)
    # voltage
    @variable(model, v[1:data[:e]])
    # DER reactive power injection
    @variable(model, q[1:data[:e]])
    @variable(model, v_slack[1:data[:e]]>=0)
    # objective function value 
    @objective(model, Min, (q.-q_current)'*(q.-q_current) + 10000*v_slack'*v_slack)
    # DER injection limits 
    @constraint(model, q .<= data[:q̅])
    @constraint(model, q .>= data[:q̲])
    # voltage limits 
    @constraint(model, v .>= data[:v̲] .- v_slack)
    @constraint(model, v .<= data[:v̅] .+ v_slack)
    # voltage definition
    @constraint(model, v .== 1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q))
    # solve model
    optimize!(model)
    @info("Distribution OPF terminates with status $(termination_status(model))")
    return Dict(:v => JuMP.value.(v), :Δq => JuMP.value.(q).-q_current, :obj => JuMP.objective_value(model))
end


# number of simulation time steps 
T = 1000 

# synthesize active power load profile 
# (at t₁, storage injection drops by 50%. At t₂, storage restores the original injection)
t₁ = 300 
t₂ = 700
p = data[:d_p] .* ones(T)'
for t in 1:T 
    t <  t₁ ? p[3,t] = data[:d_p][3] : NaN 
    t >= t₁ && t < t₂ ? p[3,t] = data[:d_p][3] * 0.5 : NaN 
    t >= t₂ ? p[3,t] = data[:d_p][3] : NaN 
end
p += rand(Normal(0,0.002),3,T) 

# droop control activates at time step t₃
t₃ = 100


###### Droop controller 
v = zeros(3,T)
q_inv = zeros(3,T)
q = data[:d_q] 
η = 0.1 # gain coefficient of the droop controller 
for t in 1:T
    t >= t₃ ? q_inv[:,t] = clamp.(q_inv[:,t-1] .+ η * [droop(1,v[:,t-1]);droop(2,v[:,t-1]);droop(3,v[:,t-1])],data[:q̲],data[:q̅])  : NaN
    v[:,t] = voltage(p[:,t],q.+q_inv[:,t])
end

plo_volt = plot(ylims=(0.9,1.1),xlims=(0,1000),frame=:box,xlabel="time step [sec]",ylabel="voltage [p.u.]")
hline!([0.95],fillrange = 1.05, c=:gray,label=false, opacity=0.25)
plot!(sqrt.(v)',label=["bus 1" "bus 2" "bus 3"],c=[:red :green :blue],lw=2,legend=:bottomright)


plo_apwr = plot(ylims=(-0.16,0.125),xlims=(0,1000),frame=:box,xlabel="time step [sec]",ylabel="active power [kW]")
plot!(p',label=["bus 1" "bus 2" "bus 3"],c=[:red :green :blue],lw=2,legend=false)

plo_rpwr = plot(ylims=(minimum(data[:q̲])-0.02,maximum(data[:q̅])+0.02),xlims=(0,1000),frame=:box,xlabel="time step [sec]",ylabel="reactive power [kW]")
plot!(q_inv',label=["bus 1" "bus 2" "bus 3"],c=[:red :green :blue],lw=2,legend=false)

plot_droop = plot(plo_apwr,plo_rpwr,plo_volt,layout=(3,1),size=(600,600))



###### OPF-based controller 
v = zeros(3,T)
q_inv = zeros(3,T)
q = data[:d_q] 
η = 0.1 # gain coefficient of the droop controller 
for t in 1:T
    data_copy = deepcopy(data)
    data_copy[:d_p] = p[:,t]
    if t >= t₃
        sol_opf = DistOPF(data_copy,q_inv[:,t-1])
        q_inv[:,t] = clamp.(q_inv[:,t-1] .+ η * sol_opf[:Δq],data[:q̲],data[:q̅])
    end
    v[:,t] = voltage(p[:,t],q.+q_inv[:,t])
end

plo_volt = plot(ylims=(0.9,1.1),xlims=(0,1000),frame=:box,xlabel="time step [sec]",ylabel="voltage [p.u.]")
hline!([0.95],fillrange = 1.05, c=:gray,label=false, opacity=0.25)
plot!(sqrt.(v)',label=["bus 1" "bus 2" "bus 3"],c=[:red :green :blue],lw=2,legend=:bottomright)


plo_apwr = plot(ylims=(-0.16,0.125),xlims=(0,1000),frame=:box,xlabel="time step [sec]",ylabel="active power [kW]")
plot!(p',label=["bus 1" "bus 2" "bus 3"],c=[:red :green :blue],lw=2,legend=false)

plo_rpwr = plot(ylims=(minimum(data[:q̲])-0.02,maximum(data[:q̅])+0.02),xlims=(0,1000),frame=:box,xlabel="time step [sec]",ylabel="reactive power [kW]")
plot!(q_inv',label=["bus 1" "bus 2" "bus 3"],c=[:red :green :blue],lw=2,legend=false)

plot_opf = plot(plo_apwr,plo_rpwr,plo_volt,layout=(3,1),size=(600,600))

# savefig(plot_opf, "acopf_control.pdf") 



# data_copy = deepcopy(data)
# # data_copy[:d_p] = p[:,t]
# sol_opf = DistOPF(data_copy)